@@ -30,9 +30,7 @@ import org.slf4j.LoggerFactory;
 
 import org.apache.cassandra.auth.Permission;
 import org.apache.cassandra.config.*;
-import org.apache.cassandra.cql3.CFName;
-import org.apache.cassandra.cql3.CQL3Type;
-import org.apache.cassandra.cql3.ColumnIdentifier;
+import org.apache.cassandra.cql3.*;
 import org.apache.cassandra.db.ColumnFamilyStore;
 import org.apache.cassandra.db.Keyspace;
 import org.apache.cassandra.db.marshal.AbstractType;
@@ -68,29 +66,23 @@ public class AlterTableStatement extends SchemaAlteringStatement
     }
 
     public final Type oType;
-    public final CQL3Type.Raw validator;
-    public final ColumnIdentifier.Raw rawColumnName;
     private final TableAttributes attrs;
-    private final Map<ColumnIdentifier.Raw, ColumnIdentifier> renames;
-    private final boolean isStatic; // Only for ALTER ADD
+    private final Map<ColumnDefinition.Raw, ColumnDefinition.Raw> renames;
+    private final List<AlterTableStatementColumn> colNameList;
     private final Long deleteTimestamp;
 
     public AlterTableStatement(CFName name,
                                Type type,
-                               ColumnIdentifier.Raw columnName,
-                               CQL3Type.Raw validator,
+                               List<AlterTableStatementColumn> colDataList,
                                TableAttributes attrs,
-                               Map<ColumnIdentifier.Raw, ColumnIdentifier> renames,
-                               boolean isStatic,
+                               Map<ColumnDefinition.Raw, ColumnDefinition.Raw> renames,
                                Long deleteTimestamp)
     {
         super(name);
         this.oType = type;
-        this.rawColumnName = columnName;
-        this.validator = validator; // used only for ADD/ALTER commands
+        this.colNameList = colDataList;
         this.attrs = attrs;
         this.renames = renames;
-        this.isStatic = isStatic;
         this.deleteTimestamp = deleteTimestamp;
     }
 
@@ -111,15 +103,11 @@ public class AlterTableStatement extends SchemaAlteringStatement
             throw new InvalidRequestException("Cannot use ALTER TABLE on Materialized View");
 
         CFMetaData cfm;
-
-        CQL3Type validator = this.validator == null ? null : this.validator.prepare(keyspace());
         ColumnIdentifier columnName = null;
         ColumnDefinition def = null;
-        if (rawColumnName != null)
-        {
-            columnName = rawColumnName.prepare(meta);
-            def = meta.getColumnDefinition(columnName);
-        }
+        CQL3Type.Raw dataType = null;
+        boolean isStatic = false;
+        CQL3Type validator = null;
 
         List<ViewDefinition> viewUpdates = null;
         Iterable<ViewDefinition> views = View.findAll(keyspace(), columnFamily());
@@ -127,167 +115,189 @@ public class AlterTableStatement extends SchemaAlteringStatement
         switch (oType)
         {
             case ALTER:
-                // We do not support altering of types and only allow this to for people who have already one
-                // through the upgrade of 2.x CQL-created SSTables with Thrift writes, affected by CASSANDRA-15778.
-                if (meta.isDense()
-                    && meta.compactValueColumn().equals(def)
-                    && meta.compactValueColumn().type instanceof EmptyType
-                    && validator != null)
+                cfm = null;
+                for (AlterTableStatementColumn colData : colNameList)
                 {
-                    if (validator.getType() instanceof BytesType)
-                    {
-                        cfm = meta.copyWithNewCompactValueType(validator.getType());
-                        break;
-                    }
+                    columnName = colData.getColumnName().getIdentifier(meta);
+                    def = meta.getColumnDefinition(columnName);
+                    dataType = colData.getColumnType();
+                    validator = dataType.prepare(keyspace());
 
-                    throw new InvalidRequestException(String.format("Compact value type can only be changed to BytesType, but %s was given.",
-                                                                    validator.getType()));
+                    // We do not support altering of types and only allow this to for people who have already one
+                    // through the upgrade of 2.x CQL-created SSTables with Thrift writes, affected by CASSANDRA-15778.
+                    if (meta.isDense()
+                        && meta.compactValueColumn().equals(def)
+                        && meta.compactValueColumn().type instanceof EmptyType
+                        && validator != null)
+                    {
+                        if (validator.getType() instanceof BytesType)
+                            cfm = meta.copyWithNewCompactValueType(validator.getType());
+                        else
+                            throw new InvalidRequestException(String.format("Compact value type can only be changed to BytesType, but %s was given.",
+                                                                            validator.getType()));
+                    }
                 }
-                else
-                {
+
+                if (cfm == null)
                     throw new InvalidRequestException("Altering of types is not allowed");
-                }
+                else
+                    break;
             case ADD:
-                assert columnName != null;
-                if (meta.isCompactTable())
+                if (meta.isDense())
                     throw new InvalidRequestException("Cannot add new column to a COMPACT STORAGE table");
 
                 cfm = meta.copy();
 
-                if (isStatic)
+                for (AlterTableStatementColumn colData : colNameList)
                 {
-                    if (!cfm.isCompound())
-                        throw new InvalidRequestException("Static columns are not allowed in COMPACT STORAGE tables");
-                    if (cfm.clusteringColumns().isEmpty())
-                        throw new InvalidRequestException("Static columns are only useful (and thus allowed) if the table has at least one clustering column");
-                }
+                    columnName = colData.getColumnName().getIdentifier(cfm);
+                    def = cfm.getColumnDefinition(columnName);
+                    dataType = colData.getColumnType();
+                    assert dataType != null;
+                    isStatic = colData.getStaticType();
+                    validator = dataType.prepare(keyspace());
 
-                if (def != null)
-                {
-                    switch (def.kind)
+
+                    if (isStatic)
                     {
-                        case PARTITION_KEY:
-                        case CLUSTERING:
-                            throw new InvalidRequestException(String.format("Invalid column name %s because it conflicts with a PRIMARY KEY part", columnName));
-                        default:
-                            throw new InvalidRequestException(String.format("Invalid column name %s because it conflicts with an existing column", columnName));
-                    }
-                }
-
-                AbstractType<?> type = validator.getType();
-                if (type.isCollection() && type.isMultiCell())
-                {
-                    if (!cfm.isCompound())
-                        throw new InvalidRequestException("Cannot use non-frozen collections in COMPACT STORAGE tables");
-                    if (cfm.isSuper())
-                        throw new InvalidRequestException("Cannot use non-frozen collections with super column families");
-                }
-
-                ColumnDefinition toAdd = isStatic
-                                       ? ColumnDefinition.staticDef(cfm, columnName.bytes, type)
-                                       : ColumnDefinition.regularDef(cfm, columnName.bytes, type);
-
-                CFMetaData.DroppedColumn droppedColumn = meta.getDroppedColumns().get(columnName.bytes);
-                if (null != droppedColumn)
-                {
-                    if (droppedColumn.kind != toAdd.kind)
-                    {
-                        String message =
-                            String.format("Cannot re-add previously dropped column '%s' of kind %s, incompatible with previous kind %s",
-                                          columnName,
-                                          toAdd.kind,
-                                          droppedColumn.kind == null ? "UNKNOWN" : droppedColumn.kind);
-                        throw new InvalidRequestException(message);
+                        if (!cfm.isCompound())
+                            throw new InvalidRequestException("Static columns are not allowed in COMPACT STORAGE tables");
+                        if (cfm.clusteringColumns().isEmpty())
+                            throw new InvalidRequestException("Static columns are only useful (and thus allowed) if the table has at least one clustering column");
                     }
 
-                    // After #8099, not safe to re-add columns of incompatible types - until *maybe* deser logic with dropped
-                    // columns is pushed deeper down the line. The latter would still be problematic in cases of schema races.
-                    if (!type.isValueCompatibleWith(droppedColumn.type))
+                    if (def != null)
                     {
-                        String message =
-                            String.format("Cannot re-add previously dropped column '%s' of type %s, incompatible with previous type %s",
-                                          columnName,
-                                          type.asCQL3Type(),
-                                          droppedColumn.type.asCQL3Type());
-                        throw new InvalidRequestException(message);
-                    }
-
-                    // Cannot re-add a dropped counter column. See #7831.
-                    if (meta.isCounter())
-                        throw new InvalidRequestException(String.format("Cannot re-add previously dropped counter column %s", columnName));
-                }
-
-                cfm.addColumnDefinition(toAdd);
-
-                // Adding a column to a table which has an include all view requires the column to be added to the view as well
-                if (!isStatic)
-                {
-                    for (ViewDefinition view : views)
-                    {
-                        if (view.includeAllColumns)
+                        switch (def.kind)
                         {
-                            ViewDefinition viewCopy = view.copy();
-                            viewCopy.metadata.addColumnDefinition(ColumnDefinition.regularDef(viewCopy.metadata, columnName.bytes, type));
-                            if (viewUpdates == null)
-                                viewUpdates = new ArrayList<>();
-                            viewUpdates.add(viewCopy);
+                            case PARTITION_KEY:
+                            case CLUSTERING:
+                                throw new InvalidRequestException(String.format("Invalid column name %s because it conflicts with a PRIMARY KEY part", columnName));
+                            default:
+                                throw new InvalidRequestException(String.format("Invalid column name %s because it conflicts with an existing column", columnName));
+                        }
+                    }
+
+                    AbstractType<?> type = validator.getType();
+                    if (type.isCollection() && type.isMultiCell())
+                    {
+                        if (!cfm.isCompound())
+                            throw new InvalidRequestException("Cannot use non-frozen collections in COMPACT STORAGE tables");
+                        if (cfm.isSuper())
+                            throw new InvalidRequestException("Cannot use non-frozen collections with super column families");
+                    }
+
+                    ColumnDefinition toAdd = isStatic
+                                           ? ColumnDefinition.staticDef(cfm, columnName.bytes, type)
+                                           : ColumnDefinition.regularDef(cfm, columnName.bytes, type);
+
+                    CFMetaData.DroppedColumn droppedColumn = meta.getDroppedColumns().get(columnName.bytes);
+                    if (null != droppedColumn)
+                    {
+                        if (droppedColumn.kind != toAdd.kind)
+                        {
+                            String message =
+                                String.format("Cannot re-add previously dropped column '%s' of kind %s, incompatible with previous kind %s",
+                                              columnName,
+                                              toAdd.kind,
+                                              droppedColumn.kind == null ? "UNKNOWN" : droppedColumn.kind);
+                            throw new InvalidRequestException(message);
+                        }
+                        // After #8099, not safe to re-add columns of incompatible types - until *maybe* deser logic with dropped
+                        // columns is pushed deeper down the line. The latter would still be problematic in cases of schema races.
+                        if (!type.isValueCompatibleWith(droppedColumn.type))
+                        {
+                            String message =
+                                String.format("Cannot re-add previously dropped column '%s' of type %s, incompatible with previous type %s",
+                                              columnName,
+                                              type.asCQL3Type(),
+                                              droppedColumn.type.asCQL3Type());
+                            throw new InvalidRequestException(message);
+                        }
+
+                        // Cannot re-add a dropped counter column. See #7831.
+                        if (meta.isCounter())
+                            throw new InvalidRequestException(String.format("Cannot re-add previously dropped counter column %s", columnName));
+                    }
+
+                    cfm.addColumnDefinition(toAdd);
+
+                    // Adding a column to a table which has an include all view requires the column to be added to the view as well
+                    if (!isStatic)
+                    {
+                        for (ViewDefinition view : views)
+                        {
+                            if (view.includeAllColumns)
+                            {
+                                ViewDefinition viewCopy = view.copy();
+                                viewCopy.metadata.addColumnDefinition(ColumnDefinition.regularDef(viewCopy.metadata, columnName.bytes, type));
+                                if (viewUpdates == null)
+                                    viewUpdates = new ArrayList<>();
+                                viewUpdates.add(viewCopy);
+                            }
                         }
                     }
                 }
                 break;
 
             case DROP:
-                assert columnName != null;
                 if (!meta.isCQLTable())
                     throw new InvalidRequestException("Cannot drop columns from a non-CQL3 table");
 
-                if (def == null)
-                    throw new InvalidRequestException(String.format("Column %s was not found in table %s", columnName, columnFamily()));
-
                 cfm = meta.copy();
 
-                switch (def.kind)
+                for (AlterTableStatementColumn colData : colNameList)
                 {
-                    case PARTITION_KEY:
-                    case CLUSTERING:
-                        throw new InvalidRequestException(String.format("Cannot drop PRIMARY KEY part %s", columnName));
-                    case REGULAR:
-                    case STATIC:
-                        ColumnDefinition toDelete = null;
-                        for (ColumnDefinition columnDef : cfm.partitionColumns())
-                        {
-                            if (columnDef.name.equals(columnName))
+                    columnName = colData.getColumnName().getIdentifier(cfm);
+                    def = cfm.getColumnDefinition(columnName);
+
+                    if (def == null)
+                        throw new InvalidRequestException(String.format("Column %s was not found in table %s", columnName, columnFamily()));
+
+                    switch (def.kind)
+                    {
+                        case PARTITION_KEY:
+                        case CLUSTERING:
+                            throw new InvalidRequestException(String.format("Cannot drop PRIMARY KEY part %s", columnName));
+                        case REGULAR:
+                        case STATIC:
+                            ColumnDefinition toDelete = null;
+                            for (ColumnDefinition columnDef : cfm.partitionColumns())
                             {
-                                toDelete = columnDef;
-                                break;
+                                if (columnDef.name.equals(columnName))
+                                {
+                                    toDelete = columnDef;
+                                    break;
+                                }
                             }
-                        }
-                        assert toDelete != null;
-                        cfm.removeColumnDefinition(toDelete);
-                        cfm.recordColumnDrop(toDelete, deleteTimestamp == null ? queryState.getTimestamp() : deleteTimestamp);
-                        break;
+                            assert toDelete != null;
+                            cfm.removeColumnDefinition(toDelete);
+                            cfm.recordColumnDrop(toDelete, deleteTimestamp == null ? queryState.getTimestamp() : deleteTimestamp);
+                            break;
+                    }
+
+                    // If the dropped column is required by any secondary indexes
+                    // we reject the operation, as the indexes must be dropped first
+                    Indexes allIndexes = cfm.getIndexes();
+                    if (!allIndexes.isEmpty())
+                    {
+                        ColumnFamilyStore store = Keyspace.openAndGetStore(cfm);
+                        Set<IndexMetadata> dependentIndexes = store.indexManager.getDependentIndexes(def);
+                        if (!dependentIndexes.isEmpty())
+                            throw new InvalidRequestException(String.format("Cannot drop column %s because it has " +
+                                                                            "dependent secondary indexes (%s)",
+                                                                            def,
+                                                                            dependentIndexes.stream()
+                                                                                            .map(i -> i.name)
+                                                                                            .collect(Collectors.joining(","))));
+                    }
+
+                    if (!Iterables.isEmpty(views))
+                        throw new InvalidRequestException(String.format("Cannot drop column %s on base table %s with materialized views.",
+                                                                        columnName.toString(),
+                                                                        columnFamily()));
                 }
 
-                // If the dropped column is required by any secondary indexes
-                // we reject the operation, as the indexes must be dropped first
-                Indexes allIndexes = cfm.getIndexes();
-                if (!allIndexes.isEmpty())
-                {
-                    ColumnFamilyStore store = Keyspace.openAndGetStore(cfm);
-                    Set<IndexMetadata> dependentIndexes = store.indexManager.getDependentIndexes(def);
-                    if (!dependentIndexes.isEmpty())
-                        throw new InvalidRequestException(String.format("Cannot drop column %s because it has " +
-                                                                        "dependent secondary indexes (%s)",
-                                                                        def,
-                                                                        dependentIndexes.stream()
-                                                                                        .map(i -> i.name)
-                                                                                        .collect(Collectors.joining(","))));
-                }
-
-                if (!Iterables.isEmpty(views))
-                    throw new InvalidRequestException(String.format("Cannot drop column %s on base table %s with materialized views.",
-                                                                    columnName.toString(),
-                                                                    columnFamily()));
                 break;
             case DROP_COMPACT_STORAGE:
 
@@ -328,11 +338,10 @@ public class AlterTableStatement extends SchemaAlteringStatement
             case RENAME:
                 cfm = meta.copy();
 
-                for (Map.Entry<ColumnIdentifier.Raw, ColumnIdentifier> entry : renames.entrySet())
+                for (Map.Entry<ColumnDefinition.Raw, ColumnDefinition.Raw> entry : renames.entrySet())
                 {
-                    ColumnIdentifier from = entry.getKey().prepare(cfm);
-                    ColumnIdentifier to = entry.getValue();
-
+                    ColumnIdentifier from = entry.getKey().getIdentifier(cfm);
+                    ColumnIdentifier to = entry.getValue().getIdentifier(cfm);
                     cfm.renameColumn(from, to);
 
                     // If the view includes a renamed column, it must be renamed in the view table and the definition.
@@ -341,8 +350,8 @@ public class AlterTableStatement extends SchemaAlteringStatement
                         if (!view.includes(from)) continue;
 
                         ViewDefinition viewCopy = view.copy();
-                        ColumnIdentifier viewFrom = entry.getKey().prepare(viewCopy.metadata);
-                        ColumnIdentifier viewTo = entry.getValue();
+                        ColumnIdentifier viewFrom = entry.getKey().getIdentifier(viewCopy.metadata);
+                        ColumnIdentifier viewTo = entry.getValue().getIdentifier(viewCopy.metadata);
                         viewCopy.renameColumn(viewFrom, viewTo);
 
                         if (viewUpdates == null)
@@ -394,7 +403,6 @@ public class AlterTableStatement extends SchemaAlteringStatement
 
             try
             {
-                // Note: 'storeRows' is basically a marker for 3.0+ sstables.
                 boolean has2xSStables = onComma.splitToList(sstableVersionsString)
                                                .stream()
                                                .map(VersionAndType::fromString)
@@ -431,10 +439,8 @@ public class AlterTableStatement extends SchemaAlteringStatement
 
     public String toString()
     {
-        return String.format("AlterTableStatement(name=%s, type=%s, column=%s, validator=%s)",
+        return String.format("AlterTableStatement(name=%s, type=%s)",
                              cfName,
-                             oType,
-                             rawColumnName,
-                             validator);
+                             oType);
     }
 }
