@@ -51,10 +51,12 @@ import org.apache.cassandra.concurrent.StageManager;
 import org.apache.cassandra.config.DatabaseDescriptor;
 import org.apache.cassandra.config.Schema;
 import org.apache.cassandra.db.Mutation;
+import org.apache.cassandra.exceptions.RequestFailureReason;
 import org.apache.cassandra.gms.ApplicationState;
 import org.apache.cassandra.gms.EndpointState;
 import org.apache.cassandra.gms.FailureDetector;
 import org.apache.cassandra.gms.Gossiper;
+import org.apache.cassandra.gms.IEndpointStateChangeSubscriber;
 import org.apache.cassandra.gms.VersionedValue;
 import org.apache.cassandra.net.IAsyncCallbackWithFailure;
 import org.apache.cassandra.net.MessageIn;
@@ -252,12 +254,14 @@ public class MigrationCoordinator
             return false;
         }
 
-        if (Schema.instance.getVersion().equals(version))
+        if (Schema.instance.isSameVersion(version))
         {
             logger.debug("Not pulling schema for version {}, because schema versions match: " +
-                         "local={}, remote={}",
-                         version, Schema.instance.getVersion(),
-                         version);
+                         "local/real={}, local/compatible={}, remote={}",
+                         version,
+                         Schema.schemaVersionToString(Schema.instance.getRealVersion()),
+                         Schema.schemaVersionToString(Schema.instance.getAltVersion()),
+                         Schema.schemaVersionToString(version));
             return false;
         }
         return true;
@@ -317,10 +321,11 @@ public class MigrationCoordinator
         {
             // If we think we may be bootstrapping or have recently started, submit MigrationTask immediately
             logger.debug("Immediately submitting migration task for {}, " +
-                         "schema versions: local={}, remote={}",
+                         "schema versions: local/real={}, local/compatible={}, remote={}",
                          endpoint,
-                         Schema.instance.getVersion(),
-                         version);
+                         Schema.schemaVersionToString(Schema.instance.getRealVersion()),
+                         Schema.schemaVersionToString(Schema.instance.getAltVersion()),
+                         Schema.schemaVersionToString(version));
             return true;
         }
         return false;
@@ -329,7 +334,7 @@ public class MigrationCoordinator
     @VisibleForTesting
     protected boolean isLocalVersion(UUID version)
     {
-        return Schema.instance.getVersion().equals(version);
+        return Schema.instance.isSameVersion(version);
     }
 
     /**
@@ -372,12 +377,12 @@ public class MigrationCoordinator
         if (state == null)
             return FINISHED_FUTURE;
 
-        VersionedValue version = state.getApplicationState(ApplicationState.SCHEMA);
+        UUID version = state.getSchemaVersion();
 
         if (version == null)
             return FINISHED_FUTURE;
 
-        return reportEndpointVersion(endpoint, UUID.fromString(version.value));
+        return reportEndpointVersion(endpoint, version);
     }
 
     private synchronized void removeEndpointFromVersion(InetAddress endpoint, UUID version)
@@ -453,7 +458,7 @@ public class MigrationCoordinator
             this.info = info;
         }
 
-        public void onFailure(InetAddress from)
+        public void onFailure(InetAddress from, RequestFailureReason failureReason)
         {
             fail();
         }
@@ -542,7 +547,7 @@ public class MigrationCoordinator
     public boolean awaitSchemaRequests(long waitMillis)
     {
         if (!FBUtilities.getBroadcastAddress().equals(InetAddress.getLoopbackAddress()))
-            CassandraDaemon.waitForGossipToSettle();
+            Gossiper.waitToSettle();
 
         WaitQueue.Signal signal = null;
         try
