@@ -47,6 +47,7 @@ import org.apache.cassandra.db.ColumnFamilyStore;
 import org.apache.cassandra.db.Keyspace;
 import org.apache.cassandra.dht.IPartitioner;
 import org.apache.cassandra.dht.Token;
+import org.apache.cassandra.distributed.Constants;
 import org.apache.cassandra.distributed.api.ConsistencyLevel;
 import org.apache.cassandra.distributed.api.Feature;
 import org.apache.cassandra.distributed.api.ICluster;
@@ -65,6 +66,7 @@ import org.apache.cassandra.distributed.api.TokenSupplier;
 import org.apache.cassandra.distributed.shared.AbstractBuilder;
 import org.apache.cassandra.distributed.shared.InstanceClassLoader;
 import org.apache.cassandra.distributed.shared.MessageFilters;
+import org.apache.cassandra.distributed.shared.Metrics;
 import org.apache.cassandra.distributed.shared.NetworkTopology;
 import org.apache.cassandra.distributed.shared.Shared;
 import org.apache.cassandra.distributed.shared.ShutdownException;
@@ -101,7 +103,7 @@ import org.reflections.Reflections;
  */
 public abstract class AbstractCluster<I extends IInstance> implements ICluster<I>, AutoCloseable
 {
-    public static Versions.Version CURRENT_VERSION = new Versions.Version(FBUtilities.getReleaseVersionString(), Versions.getClassPath());;
+    public static Versions.Version CURRENT_VERSION = new Versions.Version(FBUtilities.getReleaseVersionString(), Versions.getClassPath());
 
     // WARNING: we have this logger not (necessarily) for logging, but
     // to ensure we have instantiated the main classloader's LoggerFactory (and any LogbackStatusListener)
@@ -148,7 +150,7 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
         private volatile Versions.Version version;
         private volatile boolean isShutdown = true;
 
-        protected IInvokableInstance delegate()
+        public IInvokableInstance delegate()
         {
             if (delegate == null)
                 throw new IllegalStateException("Can't use shut down instances, delegate is null");
@@ -225,6 +227,14 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
             throw new IllegalStateException("Cannot get live member count on shutdown instance");
         }
 
+        public Metrics metrics()
+        {
+            if (isShutdown)
+                throw new IllegalStateException();
+
+            return delegate.metrics();
+        }
+
         public NodeToolResult nodetoolResult(boolean withNotifications, String... commandAndArgs)
         {
             return delegate().nodetoolResult(withNotifications, commandAndArgs);
@@ -330,7 +340,7 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
         NetworkTopology topology = NetworkTopology.build(ipPrefix, broadcastPort, nodeIdTopology);
 
         InstanceConfig config = InstanceConfig.generate(nodeNum, ipAddress, topology, root, String.valueOf(token), seedIp, datadirCount);
-        config.set("dtest.api.cluster_id", clusterId.toString());
+        config.set(Constants.KEY_DTEST_API_CLUSTER_ID, clusterId.toString());
         if (configUpdater != null)
             configUpdater.accept(config);
 
@@ -482,14 +492,20 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
     private void schemaChange(String query, boolean ignoreStoppedInstances)
     {
         I instance = ignoreStoppedInstances ? getFirstRunningInstance() : get(1);
+        schemaChange(query, ignoreStoppedInstances, instance);
+    }
 
+    public void schemaChange(String query, boolean ignoreStoppedInstances, I instance)
+    {
         instance.sync(() -> {
             try (SchemaChangeMonitor monitor = new SchemaChangeMonitor())
             {
-                // execute the schema change
-                instance.coordinator().execute(query, ConsistencyLevel.ALL);
                 if (ignoreStoppedInstances)
                     monitor.ignoreStoppedInstances();
+                monitor.startPolling();
+
+                // execute the schema change
+                instance.coordinator().execute(query, ConsistencyLevel.ALL);
                 monitor.waitForCompletion();
             }
         }).run();
@@ -525,7 +541,7 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
         private final long timeOut;
         private final TimeUnit timeoutUnit;
         protected Predicate<IInstance> instanceFilter;
-        volatile boolean changed;
+        volatile boolean initialized;
 
         public ChangeMonitor(long timeOut, TimeUnit timeoutUnit)
         {
@@ -543,7 +559,7 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
 
         protected void signal()
         {
-            if (changed && isCompleted())
+            if (initialized && !completed.isSignaled() && isCompleted())
                 completed.signalAll();
         }
 
@@ -556,21 +572,20 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
 
         public void waitForCompletion()
         {
-            startPolling();
-            changed = true;
+            initialized = true;
             signal();
             try
             {
                 if (!completed.await(timeOut, timeoutUnit))
-                    throw new InterruptedException();
+                    throw new IllegalStateException(getMonitorTimeoutMessage());
             }
             catch (InterruptedException e)
             {
-                throw new IllegalStateException(getMonitorTimeoutMessage());
+                throw new IllegalStateException("Caught exception while waiting for completion", e);
             }
         }
 
-        private void startPolling()
+        protected void startPolling()
         {
             instances.stream().filter(instanceFilter).forEach(instance -> cleanup.add(startPolling(instance)));
         }
@@ -612,7 +627,8 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
 
         protected String getMonitorTimeoutMessage()
         {
-            return "Schema agreement not reached";
+            return String.format("Schema agreement not reached. Schema versions of the instances: %s",
+                                 instances.stream().map(IInstance::schemaVersion).collect(Collectors.toList()));
         }
     }
 
@@ -645,6 +661,8 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
         Thread.setDefaultUncaughtExceptionHandler(this::uncaughtExceptions);
         try (AllMembersAliveMonitor monitor = new AllMembersAliveMonitor())
         {
+            monitor.startPolling();
+
             // Start any instances with auto_bootstrap enabled first, and in series to avoid issues
             // with multiple nodes bootstrapping with consistent range movement enabled,
             // and then start any instances with it disabled in parallel.
@@ -690,7 +708,7 @@ public abstract class AbstractCluster<I extends IInstance> implements ICluster<I
     {
         this.ignoreUncaughtThrowable = ignoreUncaughtThrowable;
     }
-
+    
     @Override
     public void close()
     {
