@@ -18,8 +18,6 @@
 
 package org.apache.cassandra.distributed.impl;
 
-import java.io.ByteArrayInputStream;
-import java.io.DataInputStream;
 import java.io.File;
 import java.io.IOException;
 import java.net.InetAddress;
@@ -46,27 +44,29 @@ import javax.management.ListenerNotFoundException;
 import javax.management.Notification;
 import javax.management.NotificationListener;
 
+import org.apache.cassandra.batchlog.BatchlogManager;
 import org.apache.cassandra.concurrent.ScheduledExecutors;
 import org.apache.cassandra.concurrent.SharedExecutorPool;
 import org.apache.cassandra.concurrent.StageManager;
 import org.apache.cassandra.config.Config;
 import org.apache.cassandra.config.DatabaseDescriptor;
 import org.apache.cassandra.config.Schema;
+import org.apache.cassandra.config.SchemaConstants;
 import org.apache.cassandra.config.YamlConfigurationLoader;
 import org.apache.cassandra.cql3.CQLStatement;
 import org.apache.cassandra.cql3.QueryOptions;
 import org.apache.cassandra.cql3.QueryProcessor;
 import org.apache.cassandra.cql3.statements.ParsedStatement;
-import org.apache.cassandra.db.BatchlogManager;
 import org.apache.cassandra.db.ColumnFamilyStore;
-import org.apache.cassandra.db.HintedHandOffManager;
 import org.apache.cassandra.db.Keyspace;
 import org.apache.cassandra.db.Memtable;
 import org.apache.cassandra.db.SystemKeyspace;
 import org.apache.cassandra.db.commitlog.CommitLog;
 import org.apache.cassandra.db.compaction.CompactionManager;
+import org.apache.cassandra.db.monitoring.ApproximateTime;
 import org.apache.cassandra.dht.IPartitioner;
 import org.apache.cassandra.dht.Token;
+import org.apache.cassandra.distributed.Constants;
 import org.apache.cassandra.distributed.api.ICluster;
 import org.apache.cassandra.distributed.api.ICoordinator;
 import org.apache.cassandra.distributed.api.IInstance;
@@ -80,47 +80,56 @@ import org.apache.cassandra.distributed.api.SimpleQueryResult;
 import org.apache.cassandra.distributed.mock.nodetool.InternalNodeProbe;
 import org.apache.cassandra.distributed.mock.nodetool.InternalNodeProbeFactory;
 import org.apache.cassandra.distributed.shared.NetworkTopology;
+import org.apache.cassandra.distributed.shared.Metrics;
 import org.apache.cassandra.gms.ApplicationState;
 import org.apache.cassandra.gms.Gossiper;
 import org.apache.cassandra.gms.VersionedValue;
-import org.apache.cassandra.io.IVersionedSerializer;
+import org.apache.cassandra.hints.DTestSerializer;
+import org.apache.cassandra.hints.HintsService;
+import org.apache.cassandra.index.SecondaryIndexManager;
 import org.apache.cassandra.io.sstable.IndexSummaryManager;
 import org.apache.cassandra.io.sstable.format.SSTableReader;
+import org.apache.cassandra.io.util.DataInputBuffer;
 import org.apache.cassandra.io.util.DataOutputBuffer;
-import org.apache.cassandra.net.CompactEndpointSerializationHelper;
+import org.apache.cassandra.io.util.FileUtils;
+import org.apache.cassandra.metrics.CassandraMetricsRegistry;
 import org.apache.cassandra.net.IMessageSink;
 import org.apache.cassandra.net.MessageIn;
 import org.apache.cassandra.net.MessageOut;
 import org.apache.cassandra.net.MessagingService;
+import org.apache.cassandra.schema.LegacySchemaMigrator;
 import org.apache.cassandra.service.CassandraDaemon;
 import org.apache.cassandra.service.ClientState;
+import org.apache.cassandra.service.DefaultFSErrorHandler;
 import org.apache.cassandra.service.PendingRangeCalculatorService;
 import org.apache.cassandra.service.QueryState;
 import org.apache.cassandra.service.StorageService;
 import org.apache.cassandra.service.StorageServiceMBean;
 import org.apache.cassandra.streaming.StreamCoordinator;
+import org.apache.cassandra.streaming.StreamSession;
 import org.apache.cassandra.tools.Output;
 import org.apache.cassandra.tools.NodeTool;
 import org.apache.cassandra.tracing.TraceState;
 import org.apache.cassandra.tracing.Tracing;
 import org.apache.cassandra.transport.messages.ResultMessage;
+import org.apache.cassandra.utils.DiagnosticSnapshotService;
 import org.apache.cassandra.utils.ExecutorUtils;
 import org.apache.cassandra.utils.FBUtilities;
 import org.apache.cassandra.utils.JVMStabilityInspector;
-import org.apache.cassandra.utils.NanoTimeToCurrentTimeMillis;
-import org.apache.cassandra.utils.Pair;
 import org.apache.cassandra.utils.Throwables;
 import org.apache.cassandra.utils.UUIDGen;
 import org.apache.cassandra.utils.concurrent.Ref;
+import org.apache.cassandra.utils.memory.BufferPool;
 
 import static java.util.concurrent.TimeUnit.MINUTES;
 import static org.apache.cassandra.distributed.api.Feature.GOSSIP;
-import static org.apache.cassandra.distributed.api.Feature.NETWORK;
 import static org.apache.cassandra.distributed.api.Feature.NATIVE_PROTOCOL;
+import static org.apache.cassandra.distributed.api.Feature.NETWORK;
 
 public class Instance extends IsolatedExecutor implements IInvokableInstance
 {
     public final IInstanceConfig config;
+    private volatile boolean initialized = false;
 
     // should never be invoked directly, so that it is instantiated on other class loader;
     // only visible for inheritance
@@ -128,7 +137,7 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
     {
         super("node" + config.num(), classLoader);
         this.config = config;
-        Object clusterId = Objects.requireNonNull(config.get("dtest.api.cluster_id"), "cluster_id is not defined");
+        Object clusterId = Objects.requireNonNull(config.get(Constants.KEY_DTEST_API_CLUSTER_ID), "cluster_id is not defined");
         ClusterIDDefiner.setId("cluster-" + clusterId);
         InstanceIDDefiner.setInstanceId(config.num());
         FBUtilities.setBroadcastInetAddress(config.broadcastAddress().getAddress());
@@ -136,7 +145,8 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
         // Set the config at instance creation, possibly before startup() has run on all other instances.
         // setMessagingVersions below will call runOnInstance which will instantiate
         // the MessagingService and dependencies preventing later changes to network parameters.
-        Config.setOverrideLoadConfig(() -> loadConfig(config));
+        Config single = loadConfig(config);
+        Config.setOverrideLoadConfig(() -> single);
     }
 
     @Override
@@ -214,7 +224,7 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
             try
             {
                 ClientState state = ClientState.forInternalCalls();
-                state.setKeyspace(SystemKeyspace.NAME);
+                state.setKeyspace(SchemaConstants.SYSTEM_KEYSPACE_NAME);
                 QueryState queryState = new QueryState(state);
 
                 CQLStatement statement = QueryProcessor.parseStatement(query, queryState).statement;
@@ -288,7 +298,6 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
                 int fromNum = from.config().num();
                 int toNum = config().num();
 
-
                 IMessage msg = serializeMessage(message, id, from.config().broadcastAddress(), broadcastAddress());
 
                 return cluster.filters().permitInbound(fromNum, toNum, msg);
@@ -307,7 +316,12 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
             long timestamp = System.currentTimeMillis();
             out.writeInt((int) timestamp);
             messageOut.serialize(out, version);
-            return new MessageImpl(messageOut.verb.ordinal(), out.toByteArray(), id, version, from);
+            byte[] bytes = out.toByteArray();
+            if (messageOut.serializedSize(version) + 12 != bytes.length)
+                throw new AssertionError(String.format("Message serializedSize(%s) does not match what was written with serialize(out, %s) for verb %s and serializer %s; " +
+                                                       "expected %s, actual %s", version, version, messageOut.verb, messageOut.serializer.getClass(),
+                                                       messageOut.serializedSize(version) + 12, bytes.length));
+            return new MessageImpl(messageOut.verb.ordinal(), bytes, id, version, from);
         }
         catch (IOException e)
         {
@@ -315,11 +329,10 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
         }
     }
 
-    public static IMessage serializeMessage(MessageIn<?> messageIn, int id, InetSocketAddress from, InetSocketAddress to)
+    public static IMessage serializeMessage(MessageIn messageIn, int id, InetSocketAddress from, InetSocketAddress to)
     {
         try (DataOutputBuffer out = new DataOutputBuffer(1024))
         {
-            // Serialize header
             int version = MessagingService.instance().getVersion(to.getAddress());
 
             out.writeInt(MessagingService.PROTOCOL_MAGIC);
@@ -327,35 +340,12 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
             long timestamp = System.currentTimeMillis();
             out.writeInt((int) timestamp);
 
-            // Serialize the message itself
-            IVersionedSerializer serializer = MessagingService.instance().verbSerializers.get(messageIn.verb);
-            CompactEndpointSerializationHelper.serialize(from.getAddress(), out);
-
-            out.writeInt(messageIn.verb.ordinal());
-            out.writeInt(messageIn.parameters.size());
-            for (Map.Entry<String, byte[]> entry : messageIn.parameters.entrySet())
-            {
-                out.writeUTF(entry.getKey());
-                out.writeInt(entry.getValue().length);
-                out.write(entry.getValue());
-            }
-
-            if (messageIn.payload != null && serializer != MessagingService.CallbackDeterminedSerializer.instance)
-            {
-                try (DataOutputBuffer dob = new DataOutputBuffer())
-                {
-                    serializer.serialize(messageIn.payload, dob, version);
-
-                    int size = dob.getLength();
-                    out.writeInt(size);
-                    out.write(dob.getData(), 0, size);
-                }
-            }
-            else
-            {
-                out.writeInt(0);
-            }
-
+            MessageOut.serialize(out,
+                                 from.getAddress(),
+                                 messageIn.verb,
+                                 messageIn.parameters,
+                                 messageIn.payload,
+                                 version);
 
             return new MessageImpl(messageIn.verb.ordinal(), out.toByteArray(), id, version, from);
         }
@@ -388,17 +378,17 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
             {
                 UUID sessionId = UUIDGen.getUUID(ByteBuffer.wrap(sessionBytes));
                 TraceState state = Tracing.instance.get(sessionId);
-                String traceMessage = String.format("Sending %s message to %s", messageOut.verb, to);
+                String message = String.format("Sending %s message to %s", messageOut.verb, to);
                 // session may have already finished; see CASSANDRA-5668
                 if (state == null)
                 {
                     byte[] traceTypeBytes = (byte[]) messageOut.parameters.get(Tracing.TRACE_TYPE);
                     Tracing.TraceType traceType = traceTypeBytes == null ? Tracing.TraceType.QUERY : Tracing.TraceType.deserialize(traceTypeBytes[0]);
-                    TraceState.mutateWithTracing(ByteBuffer.wrap(sessionBytes), traceMessage, -1, traceType.getTTL());
+                    Tracing.instance.trace(ByteBuffer.wrap(sessionBytes), message, traceType.getTTL());
                 }
                 else
                 {
-                    state.trace(traceMessage);
+                    state.trace(message);
                     if (messageOut.verb == MessagingService.Verb.REQUEST_RESPONSE)
                         Tracing.instance.doneWithNonLocalSession(state);
                 }
@@ -418,10 +408,10 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
         }
     }
 
-    public static Pair<MessageIn<Object>, Integer> deserializeMessage(IMessage imessage)
+    public static MessageIn<Object> deserializeMessage(IMessage imessage)
     {
         // Based on org.apache.cassandra.net.IncomingTcpConnection.receiveMessage
-        try (DataInputStream input = new DataInputStream(new ByteArrayInputStream(imessage.bytes())))
+        try (DataInputBuffer input = new DataInputBuffer(imessage.bytes()))
         {
             int version = imessage.version();
             if (version > MessagingService.current_version)
@@ -437,55 +427,37 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
                 id = Integer.parseInt(input.readUTF());
             else
                 id = input.readInt();
-            if (imessage.id() != id)
-                throw new IllegalStateException(String.format("Message id mismatch: %d != %d", imessage.id(), id));
-
-            // make sure to readInt, even if cross_node_to is not enabled
-            int partial = input.readInt();
-
-            return Pair.create(MessageIn.read(input, version, id), partial);
+            long currentTime = ApproximateTime.currentTimeMillis();
+            return MessageIn.read(input, version, id, MessageIn.readConstructionTime(imessage.from().getAddress(), input, currentTime));
         }
-        catch (IOException e)
+        catch (Throwable t)
         {
-            throw new RuntimeException();
+            throw new RuntimeException(t);
         }
     }
 
     public void receiveMessage(IMessage imessage)
     {
         sync(() -> {
-            Pair<MessageIn<Object>, Integer> deserialized = null;
+            // Based on org.apache.cassandra.net.IncomingTcpConnection.receiveMessage
             try
             {
-                deserialized = deserializeMessage(imessage);
+                MessageIn message = deserializeMessage(imessage);
+                if (message == null)
+                {
+                    // callback expired; nothing to do
+                    return;
+                }
+                if (message.version <= MessagingService.current_version)
+                {
+                    MessagingService.instance().receive(message, imessage.id());
+                }
+                // else ignore message
             }
             catch (Throwable t)
             {
                 throw new RuntimeException("Exception occurred on node " + broadcastAddress(), t);
             }
-
-            MessageIn<Object> message = deserialized.left;
-            int partial = deserialized.right;
-
-            long timestamp = System.currentTimeMillis();
-            boolean isCrossNodeTimestamp = false;
-
-            if (DatabaseDescriptor.hasCrossNodeTimeout())
-            {
-                long crossNodeTimestamp = (timestamp & 0xFFFFFFFF00000000L) | (((partial & 0xFFFFFFFFL) << 2) >> 2);
-                isCrossNodeTimestamp = (timestamp != crossNodeTimestamp);
-                timestamp = crossNodeTimestamp;
-            }
-
-            if (message == null)
-            {
-                // callback expired; nothing to do
-                return;
-            }
-            if (message.version <= MessagingService.current_version)
-            {
-                MessagingService.instance().receive(message, imessage.id(), timestamp, isCrossNodeTimestamp);
-            }
         }).run();
     }
 
@@ -534,13 +506,17 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
                 assert config.networkTopology().contains(config.broadcastAddress());
                 DistributedTestSnitch.assign(config.networkTopology());
 
-                DatabaseDescriptor.setDaemonInitialized();
+                DatabaseDescriptor.daemonInitialization();
+                FileUtils.setFSErrorHandler(new DefaultFSErrorHandler());
                 DatabaseDescriptor.createAllDirectories();
 
-                // We need to persist this as soon as possible after startup checks.
+                // We need to  persist this as soon as possible after startup checks.
                 // This should be the first write to SystemKeyspace (CASSANDRA-11742)
                 SystemKeyspace.persistLocalMetadata();
+                LegacySchemaMigrator.migrate();
 
+                MessagingService.instance().verbSerializers.put(MessagingService.Verb.HINT,
+                                                                new DTestSerializer());
                 try
                 {
                     // load schema from disk
@@ -556,7 +532,7 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
                 // Replay any CommitLogSegments found on disk
                 try
                 {
-                    CommitLog.instance.recover();
+                    CommitLog.instance.recoverSegmentsOnDisk();
                 }
                 catch (IOException e)
                 {
@@ -612,11 +588,14 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
                 throw new RuntimeException(t);
             }
         }).run();
+
+        initialized = true;
     }
 
     private void mkdirs()
     {
         new File(config.getString("saved_caches_directory")).mkdirs();
+        new File(config.getString("hints_directory")).mkdirs();
         new File(config.getString("commitlog_directory")).mkdirs();
         for (String dir : (String[]) config.get("data_file_directories"))
             new File(dir).mkdirs();
@@ -625,52 +604,42 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
     private static Config loadConfig(IInstanceConfig overrides)
     {
         Map<String,Object> params = ((InstanceConfig) overrides).getParams();
-        return YamlConfigurationLoader.fromMap(params, Config.class);
+        boolean check = true;
+        if (overrides.get(Constants.KEY_DTEST_API_CONFIG_CHECK) != null)
+            check = (boolean) overrides.get(Constants.KEY_DTEST_API_CONFIG_CHECK);
+        return YamlConfigurationLoader.fromMap(params, check, Config.class);
     }
 
-    private void initializeRing(ICluster cluster)
+    public static void addToRing(boolean bootstrapping, IInstance peer)
     {
-        // This should be done outside instance in order to avoid serializing config
-        String partitionerName = config.getString("partitioner");
-        List<String> initialTokens = new ArrayList<>();
-        List<InetSocketAddress> hosts = new ArrayList<>();
-        List<UUID> hostIds = new ArrayList<>();
-        for (int i = 1 ; i <= cluster.size() ; ++i)
-        {
-            IInstanceConfig config = cluster.get(i).config();
-            initialTokens.add(config.getString("initial_token"));
-            hosts.add(config.broadcastAddress());
-            hostIds.add(config.hostId());
-        }
-
         try
         {
-            IPartitioner partitioner = FBUtilities.newPartitioner(partitionerName);
-            StorageService storageService = StorageService.instance;
-            List<Token> tokens = new ArrayList<>();
-            for (String token : initialTokens)
-                tokens.add(partitioner.getTokenFactory().fromString(token));
+            IInstanceConfig config = peer.config();
+            IPartitioner partitioner = FBUtilities.newPartitioner(config.getString("partitioner"));
+            Token token = partitioner.getTokenFactory().fromString(config.getString("initial_token"));
+            InetAddress address = config.broadcastAddress().getAddress();
 
-            for (int i = 0; i < tokens.size(); i++)
-            {
-                InetSocketAddress ep = hosts.get(i);
-                Gossiper.instance.initializeNodeUnsafe(ep.getAddress(), hostIds.get(i), 1);
-                Gossiper.instance.injectApplicationState(ep.getAddress(),
+            UUID hostId = config.hostId();
+            Gossiper.runInGossipStageBlocking(() -> {
+                Gossiper.instance.initializeNodeUnsafe(address, hostId, 1);
+                Gossiper.instance.injectApplicationState(address,
                         ApplicationState.TOKENS,
-                        new VersionedValue.VersionedValueFactory(partitioner).tokens(Collections.singleton(tokens.get(i))));
-                storageService.onChange(ep.getAddress(),
+                        new VersionedValue.VersionedValueFactory(partitioner).tokens(Collections.singleton(token)));
+                StorageService.instance.onChange(address,
                         ApplicationState.STATUS,
-                        new VersionedValue.VersionedValueFactory(partitioner).normal(Collections.singleton(tokens.get(i))));
-                Gossiper.instance.realMarkAlive(ep.getAddress(), Gossiper.instance.getEndpointStateForEndpoint(ep.getAddress()));
-                int messagingVersion = cluster.get(ep).isShutdown()
-                                       ? MessagingService.current_version
-                                       : Math.min(MessagingService.current_version, cluster.get(ep).getMessagingVersion());
-                MessagingService.instance().setVersion(ep.getAddress(), messagingVersion);
-            }
+                        bootstrapping
+                                ? new VersionedValue.VersionedValueFactory(partitioner).bootstrapping(Collections.singleton(token))
+                                : new VersionedValue.VersionedValueFactory(partitioner).normal(Collections.singleton(token)));
+                Gossiper.instance.realMarkAlive(address, Gossiper.instance.getEndpointStateForEndpoint(address));
+            });
+            int messagingVersion = peer.isShutdown()
+                    ? MessagingService.current_version
+                    : Math.min(MessagingService.current_version, peer.getMessagingVersion());
+            MessagingService.instance().setVersion(address, messagingVersion);
 
-            // check that all nodes are in token metadata
-            for (int i = 0; i < tokens.size(); ++i)
-                assert storageService.getTokenMetadata().isMember(hosts.get(i).getAddress());
+            if (!bootstrapping)
+                assert StorageService.instance.getTokenMetadata().isMember(address);
+            PendingRangeCalculatorService.instance.blockUntilFinished();
         }
         catch (Throwable e) // UnknownHostException
         {
@@ -678,6 +647,51 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
         }
     }
 
+    public static void removeFromRing(IInstance peer)
+    {
+        try
+        {
+            IInstanceConfig config = peer.config();
+            IPartitioner partitioner = FBUtilities.newPartitioner(config.getString("partitioner"));
+            Token token = partitioner.getTokenFactory().fromString(config.getString("initial_token"));
+            InetAddress address = config.broadcastAddress().getAddress();
+
+            Gossiper.runInGossipStageBlocking(() -> {
+                StorageService.instance.onChange(address,
+                        ApplicationState.STATUS,
+                        new VersionedValue.VersionedValueFactory(partitioner).left(Collections.singleton(token), 0L));
+                Gossiper.instance.removeEndpoint(address);
+            });
+            PendingRangeCalculatorService.instance.blockUntilFinished();
+        }
+        catch (Throwable e) // UnknownHostException
+        {
+            throw new RuntimeException(e);
+        }
+    }
+
+    public static void addToRingNormal(IInstance peer)
+    {
+        addToRing(false, peer);
+        assert StorageService.instance.getTokenMetadata().isMember(peer.broadcastAddress().getAddress());
+    }
+
+    public static void addToRingBootstrapping(IInstance peer)
+    {
+        addToRing(true, peer);
+    }
+
+    private static void initializeRing(ICluster cluster)
+    {
+        for (int i = 1 ; i <= cluster.size() ; ++i)
+            addToRing(false, cluster.get(i));
+
+        for (int i = 1; i <= cluster.size(); ++i)
+            assert StorageService.instance.getTokenMetadata().isMember(cluster.get(i).broadcastAddress().getAddress());
+
+        StorageService.instance.setNormalModeUnsafe();
+    }
+
     public Future<Void> shutdown()
     {
         return shutdown(true);
@@ -686,6 +700,9 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
     @Override
     public Future<Void> shutdown(boolean graceful)
     {
+        if (!graceful)
+            MessagingService.instance().shutdown(false);
+
         Future<?> future = async((ExecutorService executor) -> {
             Throwable error = null;
 
@@ -696,38 +713,36 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
             if (config.has(GOSSIP) || config.has(NETWORK))
             {
                 StorageService.instance.shutdownServer();
-
-                error = parallelRun(error, executor,
-                    () -> NanoTimeToCurrentTimeMillis.shutdown(MINUTES.toMillis(1L))
-                );
             }
 
-            error = parallelRun(error, executor,
-                                (IgnoreThrowingRunnable) MessagingService.instance()::shutdown
-            );
-
             error = parallelRun(error, executor,
                                 () -> Gossiper.instance.stopShutdownAndWait(1L, MINUTES),
                                 CompactionManager.instance::forceShutdown,
-                                () -> BatchlogManager.shutdownAndWait(1L, MINUTES),
-                                () -> HintedHandOffManager.instance.shutdownAndWait(1L, MINUTES),
+                                () -> BatchlogManager.instance.shutdownAndWait(1L, MINUTES),
+                                HintsService.instance::shutdownBlocking,
                                 () -> StreamCoordinator.shutdownAndWait(1L, MINUTES),
+                                () -> StreamSession.shutdownAndWait(1L, MINUTES),
+                                () -> SecondaryIndexManager.shutdownAndWait(1L, MINUTES),
                                 () -> IndexSummaryManager.instance.shutdownAndWait(1L, MINUTES),
                                 () -> ColumnFamilyStore.shutdownExecutorsAndWait(1L, MINUTES),
                                 () -> PendingRangeCalculatorService.instance.shutdownExecutor(1L, MINUTES),
-                                () -> StorageService.instance.shutdownBGMonitorAndWait(1L, MINUTES),
+                                () -> BufferPool.shutdownLocalCleaner(1L, MINUTES),
                                 () -> Ref.shutdownReferenceReaper(1L, MINUTES),
                                 () -> Memtable.MEMORY_POOL.shutdownAndWait(1L, MINUTES),
                                 () -> SSTableReader.shutdownBlocking(1L, MINUTES),
-                                () -> ScheduledExecutors.shutdownAndWait(1L, MINUTES)
+                                () -> DiagnosticSnapshotService.instance.shutdownAndWait(1L, MINUTES)
             );
             error = parallelRun(error, executor,
-                                CommitLog.instance::shutdownBlocking
+                                () -> ScheduledExecutors.shutdownAndWait(1L, MINUTES),
+                                (IgnoreThrowingRunnable) MessagingService.instance()::shutdown
             );
             error = parallelRun(error, executor,
                                 () -> StageManager.shutdownAndWait(1L, MINUTES),
                                 () -> SharedExecutorPool.SHARED.shutdownAndWait(1L, MINUTES)
             );
+            error = parallelRun(error, executor,
+                                CommitLog.instance::shutdownBlocking
+            );
 
             Throwables.maybeFail(error);
         }).apply(isolatedExecutor);
@@ -738,6 +753,9 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
 
     public int liveMemberCount()
     {
+        if (!initialized || isShutdown())
+            return 0;
+
         return sync(() -> {
             if (!DatabaseDescriptor.isDaemonInitialized() || !Gossiper.instance.isEnabled())
                 return 0;
@@ -745,6 +763,11 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
         }).call();
     }
 
+    public Metrics metrics()
+    {
+        return callOnInstance(() -> new InstanceMetrics(CassandraMetricsRegistry.Metrics));
+    }
+
     public NodeToolResult nodetoolResult(boolean withNotifications, String... commandAndArgs)
     {
         return sync(() -> {
@@ -923,4 +946,3 @@ public class Instance extends IsolatedExecutor implements IInvokableInstance
         }
     }
 }
-
