@@ -72,7 +72,7 @@ public class InstanceConfig implements IInstanceConfig
                            String[] data_file_directories,
                            String commitlog_directory,
                            String hints_directory,
-//                           String cdc_directory,
+                           String cdc_raw_directory,
                            String initial_token)
     {
         this.num = num;
@@ -87,7 +87,7 @@ public class InstanceConfig implements IInstanceConfig
                 .set("data_file_directories", data_file_directories)
                 .set("commitlog_directory", commitlog_directory)
                 .set("hints_directory", hints_directory)
-//                .set("cdc_directory", cdc_directory)
+                .set("cdc_raw_directory", cdc_raw_directory)
                 .set("initial_token", initial_token)
                 .set("partitioner", "org.apache.cassandra.dht.Murmur3Partitioner")
                 .set("start_native_transport", true)
@@ -251,7 +251,7 @@ public class InstanceConfig implements IInstanceConfig
                                   datadirs(datadirCount, root, nodeNum),
                                   String.format("%s/node%d/commitlog", root, nodeNum),
                                   String.format("%s/node%d/hints", root, nodeNum),
-//                                  String.format("%s/node%d/cdc", root, nodeNum),
+                                  String.format("%s/node%d/cdc", root, nodeNum),
                                   token);
     }
 
